////////////////////////////////////////////////////////////////////////
// Class:       EvaluateAdcThresholdROI
// Plugin Type: analyzer (art v3_05_01)
// File:        EvaluateAdcThresholdROI_module.cc
//
// Generated at Wed Jul  8 11:41:19 2020 by Wanwei Wu using cetskelgen
// from cetlib version v3_10_00.
// About: evaluate the efficiency and purity of AdcThresholdRoi method
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "canvas/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "art_root_io/TFileService.h"
#include "art/Utilities/make_tool.h"

#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"
#include "lardata/DetectorInfoServices/DetectorClocksService.h"
#include "larcore/Geometry/Geometry.h"
#include "lardataobj/Simulation/SimChannel.h"
#include "lardataobj/RecoBase/Wire.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusService.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusProvider.h"

#include "TTree.h"
#include "TH1D.h"
#include "TEfficiency.h"

using namespace std;

class EvaluateAdcThresholdROI;


class EvaluateAdcThresholdROI : public art::EDAnalyzer {
public:
  explicit EvaluateAdcThresholdROI(fhicl::ParameterSet const& p);
  // The compiler-generated destructor is fine for non-base
  // classes without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  EvaluateAdcThresholdROI(EvaluateAdcThresholdROI const&) = delete;
  EvaluateAdcThresholdROI(EvaluateAdcThresholdROI&&) = delete;
  EvaluateAdcThresholdROI& operator=(EvaluateAdcThresholdROI const&) = delete;
  EvaluateAdcThresholdROI& operator=(EvaluateAdcThresholdROI&&) = delete;

  // Required functions.
  void analyze(art::Event const& e) override;

private:

  // User's function
  void beginJob() override;
  void endJob() override;

  // fhicl parameters
  art::InputTag fWireProducerLabel;
  art::InputTag fSimulationProducerLabel;
  std::vector<double> fThreshold; // for ROI purity: threshold on either positive or negative adc [ADC]
  std::vector<double> fRoiThreshold; // for ROI purity: cut on positive adc [ADC]

  // TTree
  TTree *fTree;
  int run;
  int subrun;
  int event;

  // for efficiency
  TH1D *h1_energy[2];
  TH1D *h1_energy_roi[2];
  TEfficiency *eff_energy[2];

  // for efficiency with cut
  TH1D *h1_energy_roi_with_cut[2];
  TEfficiency *eff_energy_with_cut[2];



  // for purity
  TH1D *h1_roi[2];
  TH1D *h1_roi_height[2];
  TH1D *h1_roi_height_sim[2];
  TH1D *h1_roi_max[2];
  TH1D *h1_roi_max_sim[2];

};


EvaluateAdcThresholdROI::EvaluateAdcThresholdROI(fhicl::ParameterSet const& p)
  : EDAnalyzer{p}  // ,
  // More initializers here.
{
  fWireProducerLabel         = p.get<art::InputTag>("WireProducerLabel");
  fSimulationProducerLabel   = p.get<art::InputTag>("SimulationProducerLabel","largeant");
  fThreshold                 = p.get<std::vector<double>>("Threshold"); // this is not necessary; can be removed 
  fRoiThreshold              = p.get<std::vector<double>>("RoiThreshold");
}


void EvaluateAdcThresholdROI::analyze(art::Event const& e)
{
  run = e.run();
  subrun = e.subRun();
  event = e.id().event();

  auto const* geo = lar::providerFrom<geo::Geometry>();
  auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService>()->DataFor(e);
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataFor(e, clockData);
  auto const& chStatus = art::ServiceHandle< lariov::ChannelStatusService >()->GetProvider();

  art::Handle < std::vector < recob::Wire > > wireListHandle;
  std::vector < art::Ptr < recob::Wire > > wires;
  if (e.getByLabel(fWireProducerLabel, wireListHandle)) {
    art::fill_ptr_vector(wires, wireListHandle);
  }

  //cout << "wires.size(): " << wires.size() << endl;

  auto simChannelHandle = e.getValidHandle<std::vector<sim::SimChannel>>(fSimulationProducerLabel);

  //cout << "(*simChannelHandle).size(): " << (*simChannelHandle).size() << endl;

  // efficiency: according to each simulated energy deposit
  // ... loop over SimChannels
  for ( auto const& simchannel : (*simChannelHandle) ){
    // get SimChannel channel number
    const raw::ChannelID_t ch1 = simchannel.Channel();
    if (chStatus.IsBad(ch1)) continue;

    int view = geo->View(ch1);
    
    auto const& timeSlices = simchannel.TDCIDEMap();

    for (auto const& timeSlice : timeSlices) {
      auto const& energyDeposits = timeSlice.second;
      auto const tpctime = timeSlice.first;
      int tdctick = static_cast<int>(clockData.TPCTDC2Tick(double(tpctime)));
      if(tdctick<0 || tdctick>int(detProp.ReadOutWindowSize())-1) continue;

      double totalE = 0;
      for ( auto const& energyDeposit : energyDeposits ){
        totalE += energyDeposit.energy;
      }
      
      h1_energy[view]->Fill(totalE);
      
      for (auto & wire : wires) {
        if (wire->Channel() != ch1) continue;
        const recob::Wire::RegionsOfInterest_t& signalROI = wire->SignalROI();

        std::vector<float> vecADC = wire->Signal(); // a zero-padded full length vector filled with RoI signal.

        for(const auto& range : signalROI.get_ranges()){
          const auto& waveform = range.data();
          raw::TDCtick_t roiFirstBinTick = range.begin_index();
          
          bool select_ROI = false;
          // with threshold cut
          for (unsigned int k=roiFirstBinTick; k<roiFirstBinTick+waveform.size(); k++) {
            //cout << "k: " << k << " ; adc: " << vecADC[k] << endl;
            if (vecADC[k] > fRoiThreshold[view]) {
              select_ROI = true;
              break;
            }
          }

          if (tdctick>=roiFirstBinTick && tdctick < int(roiFirstBinTick + waveform.size())){
            h1_energy_roi[view]->Fill(totalE);
            if (select_ROI) h1_energy_roi_with_cut[view]->Fill(totalE);
            break;
          }
        }

      } // loop over wires
    } // loop over timeSlices
  } // loop over simChannels

  // loop over all wires for purity
  for (auto & wire : wires) {
    const raw::ChannelID_t wirechannel = wire->Channel();
    if (chStatus.IsBad(wirechannel)) continue;

    int view = wire->View();

    const recob::Wire::RegionsOfInterest_t& signalROI = wire->SignalROI();

    std::vector<float> vecADC = wire->Signal(); // a zero-padded full length vector filled with ROI signal.

    for (const auto& range : signalROI.get_ranges()) {
      
      raw::TDCtick_t roiFirstBinTick = range.begin_index();
      const auto& waveform = range.data();
 
      // mean adc for ADC values great than abs(threshold)
      int nadc_sig = 0;
      double meanadc_sig = 0.;
      double maxadc_sig = 0.;
      for (unsigned int k=roiFirstBinTick; k<roiFirstBinTick+waveform.size(); k++) {
        if (std::abs(vecADC[k]) > maxadc_sig) maxadc_sig = std::abs(vecADC[k]);
        if (std::abs(vecADC[k]) > fThreshold[view]) {
          nadc_sig++;
          meanadc_sig += std::abs(vecADC[k]);
        }
      }
      meanadc_sig = meanadc_sig/nadc_sig;
      h1_roi_height[view]->Fill(meanadc_sig);
      h1_roi_max[view]->Fill(maxadc_sig);

      h1_roi[view]->Fill(0); // total roi

      // with roi threshold
      bool isSigROI = false;
      bool selectROI = false;
      
      for (unsigned int k=roiFirstBinTick; k<roiFirstBinTick+waveform.size(); k++) {
        if (vecADC[k] > fRoiThreshold[view]) {
          selectROI = true;
          break;
        }
      }

      if (selectROI) h1_roi[view]->Fill(2); // total selected roi
      
      // check simChannels
      for (auto const& simchannel : (*simChannelHandle)) {
        if (wirechannel != simchannel.Channel()) continue;

        auto const& timeSlices = simchannel.TDCIDEMap();
        for ( auto const& timeSlice : timeSlices ){
          auto const tpctime = timeSlice.first;
          int tdctick = static_cast<int>(clockData.TPCTDC2Tick(double(tpctime)));
          if(tdctick<0 || tdctick>int(detProp.ReadOutWindowSize())-1) continue;

          if (tdctick>=roiFirstBinTick && tdctick < int(roiFirstBinTick + waveform.size())){
            isSigROI = true;
            break;
          }
        }
      } // loop over simChannels

      if (isSigROI) {
        h1_roi[view]->Fill(1); // sig roi
        h1_roi_height_sim[view]->Fill(meanadc_sig);
        h1_roi_max_sim[view]->Fill(maxadc_sig);
      }

      if (isSigROI && selectROI) h1_roi[view]->Fill(3); // selected sig roi

    } // loop over roi  ranges

  } // loop over wires

} // analyze


void EvaluateAdcThresholdROI::beginJob()
{
  art::ServiceHandle<art::TFileService const> tfs;
  fTree = tfs->make<TTree>("anatree","adc threshold roi");
  fTree->Branch("run",&run,"run/I");
  fTree->Branch("subrun",&subrun,"subrun/I");
  fTree->Branch("event",&event,"event/I");

  for (int p=0; p<2; p++) {
    h1_energy[p] = tfs->make<TH1D>(Form("h1_energy_%d", p), Form("Plane %d; Energy (MeV);", p), 100, 0, 1);
    h1_energy_roi[p] = tfs->make<TH1D>(Form("h1_energy_roi_%d", p), Form("Plane %d; Energy (MeV);", p), 100, 0, 1);

    h1_energy_roi_with_cut[p] = tfs->make<TH1D>(Form("h1_energy_roi_with_cut_%d", p), Form("Plane %d (ROI threshold > %0.f ADC); Energy (MeV);", p, fRoiThreshold[p]), 100,0,1);
    
    h1_roi[p] = tfs->make<TH1D>(Form("h1_roi_%d",p), Form("Plane %d (ROI threshold > %0.f ADC); Purity;", p, fRoiThreshold[p]), 5,0,5);

    h1_roi_height[p] = tfs->make<TH1D>(Form("h1_roi_height_%d",p), Form("Plane %d; Mean adc with ADC > %0.f [ADC];", p, fThreshold[p]), 30,0,30);
    h1_roi_height_sim[p] = tfs->make<TH1D>(Form("h1_roi_height_sim_%d",p), Form("Plane %d; Mean adc with ADC > %0.f [ADC];", p, fThreshold[p]), 30,0,30);

    h1_roi_max[p] = tfs->make<TH1D>(Form("h1_roi_max_%d",p), Form("Plane %d; Max adc with ADC > %0.f [ADC];", p, fThreshold[p]), 30,0,30);
    h1_roi_max_sim[p] = tfs->make<TH1D>(Form("h1_roi_max_sim_%d",p), Form("Plane %d; Max adc with ADC > %0.f [ADC];", p, fThreshold[p]), 30,0,30);

  }

} // beginJob


void EvaluateAdcThresholdROI::endJob()
{
  art::ServiceHandle<art::TFileService const> tfs;
  
  for (int p=0; p<2; p++) {
    if(TEfficiency::CheckConsistency(*(h1_energy_roi[p]), *(h1_energy[p]))){
      eff_energy[p] = tfs->make<TEfficiency>(*(h1_energy_roi[p]), *(h1_energy[p]));
      eff_energy[p]->Write(Form("eff_energy_%d",p));
    } 

    if(TEfficiency::CheckConsistency(*(h1_energy_roi_with_cut[p]), *(h1_energy[p]))){
      eff_energy_with_cut[p] = tfs->make<TEfficiency>(*(h1_energy_roi_with_cut[p]), *(h1_energy[p]));
      eff_energy_with_cut[p]->Write(Form("eff_energy_with_cut_%d",p));
    }
  }
} // endJob


DEFINE_ART_MODULE(EvaluateAdcThresholdROI)
